︠ef693524-bef4-4254-bad8-6aff77b26adai︠

%md

### Group theory

Recall the definition of a group:

https://en.wikipedia.org/wiki/Group_(mathematics)#Definition

There are tons and tons of groups.

Some are infinite and some are finite.

Cryptography is mostly concerned with finite groups.

︡2c836d85-f9da-4dfb-a423-f558833d2bff︡{"done":true,"md":"\n### Group theory\n\nRecall the definition of a group:\n\nhttps://en.wikipedia.org/wiki/Group_(mathematics)#Definition\n\nThere are tons and tons of groups.\n\nSome are infinite and some are finite.\n\nCryptography is mostly concerned with finite groups."}
︠385f1fc2-a297-46d7-aaad-c9bf2c3adc33i︠
%md

### Common groups in cryptography

* $\mathbb{Z}_n$ is addition modulo $n$.
* $\mathbb{Z}_n^*$ is multiplication modulo $n$ on elements coprime to $n$.
* The multiplicative part of $GF(p^k)$.
* Elliptic curve groups.
︡5e61591d-2b70-41cb-bcba-698138f5d36b︡{"done":true,"md":"\n### Common groups in cryptography\n\n* $\\mathbb{Z}_n$ is addition modulo $n$.\n* $\\mathbb{Z}_n^*$ is multiplication modulo $n$ on elements coprime to $n$. \n* The multiplicative part of $GF(p^k)$.\n* Elliptic curve groups."}
︠77f6fade-ffe6-466b-9513-97589f004b67i︠
%md

### Multiplicative or additive notation

When we talk about a generic group $G$ we will sometimes think of the operation as being multiplication.

Other times we will think of it as being more like addition.

Additive notation: $(G,+)$

* The identity is 0
* The inverse of $a$ is $-a$
* $na = \underbrace{a+a+a+\cdots+a}_{n\,times}$.
* The DLP:  Solve for $n$ in the equation: $na = b$.
︡3b3d7dda-a4a1-40b1-881c-6b7e4a41b865︡{"done":true,"md":"\n### Multiplicative or additive notation\n\nWhen we talk about a generic group $G$ we will sometimes think of the operation as being multiplication.\n\nOther times we will think of it as being more like addition.  \n\nAdditive notation: $(G,+)$\n\n* The identity is 0\n* The inverse of $a$ is $-a$\n* $na = \\underbrace{a+a+a+\\cdots+a}_{n\\,times}$.\n* The DLP:  Solve for $n$ in the equation: $na = b$."}
︠688992e7-adb8-44c0-a119-d894de238874i︠
%md

### Multiplicative or additive notation

When we talk about a generic group $G$ we will sometimes think of the operation as being multiplication.

Other times we will think of it as being more like addition.

Multiplicative notation: $(G,\cdot)$

* The identity is 1
* The inverse of $a$ is $a^{-1}$
* $a^n = \underbrace{a\cdot a \cdot a \cdots a}_{n\,times}$.
* The DLP:  Solve for $n$ in the equation: $a^n = b$.
︡87e28e03-5b71-4d7e-980a-14f977f84cc4︡{"done":true,"md":"\n### Multiplicative or additive notation\n\nWhen we talk about a generic group $G$ we will sometimes think of the operation as being multiplication.\n\nOther times we will think of it as being more like addition.  \n\nMultiplicative notation: $(G,\\cdot)$\n\n* The identity is 1\n* The inverse of $a$ is $a^{-1}$\n* $a^n = \\underbrace{a\\cdot a \\cdot a \\cdots a}_{n\\,times}$.\n* The DLP:  Solve for $n$ in the equation: $a^n = b$."}
︠df42cbd6-8d51-4d5e-b2e6-9a377d5baf90i︠
%md

### The order of an element

Given $a$ in $G$, the order of $a$ is the smallest positive integer $k$ such that $a^k = 1$.
︡7a5c9ede-1850-4bf5-a0bb-62d12805e7f8︡{"done":true,"md":"\n### The order of an element\n\nGiven $a$ in $G$, the order of $a$ is the smallest positive integer $k$ such that $a^k = 1$."}
︠556c5fcf-1b05-4888-a3d1-47bafd66f8f0︠
G = Zmod(397)
g = G.random_element()
g
︡ab250c9a-0b0b-4f3b-9ee6-da632beae16f︡{"stdout":"198\n"}︡{"done":true}
︠b2e2021b-742c-464f-a2ef-18ef87cbecf3︠
g.multiplicative_order()
︡bfc29298-2b00-4384-a322-9137de7fbb78︡{"stdout":"44\n"}︡{"done":true}
︠d3d721c9-df3f-4256-bd51-1ac4dde86245︠
for k in range(g.multiplicative_order()+1):
    print("g^{:2} = {:3}".format(k,g^k))
︡8ace977a-0b65-44b3-be85-ba576434e392︡{"stdout":"g^ 0 = 1  \ng^ 1 = 198\ng^ 2 = 298\ng^ 3 = 248\ng^ 4 = 273\ng^ 5 = 62 \ng^ 6 = 366\ng^ 7 = 214\ng^ 8 = 290\ng^ 9 = 252\ng^10 = 271\ng^11 = 63 \ng^12 = 167\ng^13 = 115\ng^14 = 141\ng^15 = 128\ng^16 = 333\ng^17 = 32 \ng^18 = 381\ng^19 = 8  \ng^20 = 393\ng^21 = 2  \ng^22 = 396\ng^23 = 199\ng^24 = 99 \ng^25 = 149\ng^26 = 124\ng^27 = 335\ng^28 = 31 \ng^29 = 183\ng^30 = 107\ng^31 = 145\ng^32 = 126\ng^33 = 334\ng^34 = 230\ng^35 = 282\ng^36 = 256\ng^37 = 269\ng^38 = 64 \ng^39 = 365\ng^40 = 16 \ng^41 = 389\ng^42 = 4  \ng^43 = 395\ng^44 = 1  \n"}︡{"done":true}
︠a992ad6b-4a59-4699-b41b-12d3ba8a28b1︠
## The order of g is the size of the set it can generate

len(set([g^k for k in range(G.order())]))
︡fe6e243a-4950-40f8-b618-ed48d5785402︡{"stdout":"44\n"}︡{"done":true}
︠8245c55c-fddd-4ce1-9d9b-e0bd82c0d9ecs︠
euler_phi(397)/44
︡0887d956-13ea-443b-9812-394ed84165c0︡{"stdout":"9\n"}︡{"done":true}
︠7e2afb1c-9a4a-495c-8353-43ae1dd3396b︠
## The order of g divides |G|

euler_phi(397)%g.multiplicative_order()
︡12997099-e3fb-49e0-a1c6-0dbaddaab09d︡{"stdout":"0\n"}︡{"done":true}
︠ce9f89af-b608-42a1-973d-c6075c1ae4a2i︠
%md

### Cyclic groups

The group $G$ is cyclic if it contains an element of order $|G|$.

In other words some single element generates  $G$.

There is some $\alpha \in G$ such that every $\beta \in G$ is $\alpha^x$ for some index $x$.

︡70467900-e764-4576-af90-926ec015ee83︡{"done":true,"md":"\n### Cyclic groups\n\nThe group $G$ is cyclic if it contains an element of order $|G|$.\n\nIn other words some single element generates  $G$.\n\nThere is some $\\alpha \\in G$ such that every $\\beta \\in G$ is $\\alpha^x$ for some index $x$."}
︠4360d51c-56bb-4431-902f-ce63438f1ac7i︠

%md
### Theorem

For every prime $p$, the group $(\mathbb{Z}_p^*,\cdot)$ is a finite cyclic group.

︡59b06294-0f45-44b2-b373-94386cd8b0a1︡{"done":true,"md":"### Theorem\n\nFor every prime $p$, the group $(\\mathbb{Z}_p^*,\\cdot)$ is a finite cyclic group."}
︠6d5b27ec-1acb-4457-96c4-78c5c793de4d︠
## Empirical check:
r = randint(0,2193821983798217)
p = next_prime(r)
G = Zmod(p)
G.multiplicative_group_is_cyclic()
︡aa07beb2-dc95-4bf2-b4c5-e3b6861c6f1f︡{"stdout":"True\n"}︡{"done":true}
︠cf62dd64-9af6-4ad6-b4aa-6ebcc527d33b︠
## Not true for all composite numbers:

Zmod(8).multiplicative_group_is_cyclic()
︡ca8820c2-b8d5-423f-b93d-c6d286e89325︡{"stdout":"False\n"}︡{"done":true}
︠c2ec59d9-e45b-4893-812a-b727e53c55aes︠

def show_orders(n):
    """Print the orders of the elements of Z_n^*"""
    print("The multiplicative orders of the elements of Z_{}".format(n))
    for g in Zmod(n):
        if gcd(g,n) != 1:
            continue
        print("the order of {} is {}".format(g,g.multiplicative_order()))

show_orders(19)
︡04b320d8-c23e-471d-8c7a-67580e95e334︡{"stdout":"The multiplicative orders of the elements of Z_19\nthe order of 1 is 1\nthe order of 2 is 18\nthe order of 3 is 18\nthe order of 4 is 9\nthe order of 5 is 9\nthe order of 6 is 9\nthe order of 7 is 3\nthe order of 8 is 6\nthe order of 9 is 9\nthe order of 10 is 18\nthe order of 11 is 3\nthe order of 12 is 6\nthe order of 13 is 18\nthe order of 14 is 18\nthe order of 15 is 18\nthe order of 16 is 9\nthe order of 17 is 9\nthe order of 18 is 2\n"}︡{"done":true}
︠4e107562-46b9-4617-aad6-d1470fbf81dfi︠
%md

### [Theorem](https://en.wikipedia.org/wiki/Lagrange%27s_theorem_(group_theory))

For every element $g$ of a finite group $G$, the order of $g$ divides $|G|$.

#### Proof:
Let $g \in G$ be given:

$\displaystyle \prod_{a \in G} a = \prod_{a \in G} ga = g^{|G|} \prod_{a\in G} a$

Therefore

$1 = g^{|G|}$

This implies that the order of $g$ divides $|G|$.

It also generalizes FLT to finite groups.

︡3c00ca23-52c1-4c37-9b8f-3d8efc578872︡{"done":true,"md":"\n### [Theorem](https://en.wikipedia.org/wiki/Lagrange%27s_theorem_(group_theory))\n\nFor every element $g$ of a finite group $G$, the order of $g$ divides $|G|$.\n\n#### Proof:\nLet $g \\in G$ be given:\n\n$\\displaystyle \\prod_{a \\in G} a = \\prod_{a \\in G} ga = g^{|G|} \\prod_{a\\in G} a$\n\nTherefore\n\n$1 = g^{|G|}$\n\nThis implies that the order of $g$ divides $|G|$.\n\nIt also generalizes FLT to finite groups."}
︠6df9c990-76e4-45a1-aac5-a9f2ee39f770i︠
%md

### Theorem

Let $G$ be a finite cyclic group.  Then

1. The number of primitive elements of $G$ is $\varphi(|G|)$.
2. If $|G|$ is prime, then all elements $a \neq 1$ in $G$ are primitive.
︡96e8c441-f2e2-4764-9579-21a46efa61d5︡{"done":true,"md":"\n### Theorem\n\nLet $G$ be a finite cyclic group.  Then\n\n1. The number of primitive elements of $G$ is $\\varphi(|G|)$.\n2. If $|G|$ is prime, then all elements $a \\neq 1$ in $G$ are primitive."}
︠6717fea1-c1d6-404a-a96b-d62dbfa7e11a︠
p = next_prime(293847)
G = Zmod(p)
generators = [g for g in G if g != 0 and g.multiplicative_order()==p-1]
len(generators) == euler_phi(p-1)
︡62c5eb15-1fb3-43c4-8f9c-06949ac11b45︡{"stdout":"True\n"}︡{"done":true}
︠e640b56a-dc4a-429b-ab57-71acbf4b699di︠
%md

Note that if $p$ is prime then $|\mathbb{Z}_p^*| = p-1$ is never prime.
︡acbc529e-bbc1-44b7-ae3b-50cb2c0e36db︡{"done":true,"md":"\nNote that if $p$ is prime then $|\\mathbb{Z}_p^*| = p-1$ is never prime."}
︠63b9b74e-4b10-40e0-9e72-162044d3fca3i︠
%md

But when $p$ is prime and $G=\mathbb{Z}_p$, then $|G|=p$, and any element is a generator.
︡21b1f540-c25d-4b52-842c-aa58ea81060d︡{"done":true,"md":"\nBut when $p$ is prime and $G=\\mathbb{Z}_p$, then $|G|=p$, and any element is a generator."}
︠c777d1b7-a87b-4ef7-972e-d9aa90b16f3cs︠
# Any element is a generator of the additive group Z_p
n=7
r = randint(1,7)
for g in range(1,7):
    print((g*r)%n)
︡5c7f10ad-40b3-465d-a3ce-a4f0192300db︡{"stdout":"4\n1\n5\n2\n6\n3\n"}︡{"done":true}
︠6d5ecb65-d814-4bd0-9ba3-4eccbe650b36i︠
%md

### Subgroups

A **subgroup** is a subset of a group that is closed under the group operation.

#### Example of a subgroup

The powers of $7$ are a subgroup of the group $\mathbb{Z}_n^*$, whenever $\gcd(7,n)=1$.

*Check closure*:

Let $7^j$ and $7^k$ be two powers of 7 mod $n$ for integers $j,k$.

Then

$7^j\cdot 7^k \equiv 7^{j+k} \mod n$.

But $7^{j+k}$ is obviously a power of $7$.

Therefore the powers of 7 are closed under the group operation.

Therefore the powers of 7 form a subgroup.
︡084a075d-e9f0-465e-b825-db1dda12f64e︡{"done":true,"md":"\n### Subgroups\n\nA **subgroup** is a subset of a group that is closed under the group operation.\n\n#### Example of a subgroup\n\nThe powers of $7$ are a subgroup of the group $\\mathbb{Z}_n^*$, whenever $\\gcd(7,n)=1$.  \n\n*Check closure*:\n\nLet $7^j$ and $7^k$ be two powers of 7 mod $n$ for integers $j,k$.\n\nThen\n\n$7^j\\cdot 7^k \\equiv 7^{j+k} \\mod n$.\n\nBut $7^{j+k}$ is obviously a power of $7$.\n\nTherefore the powers of 7 are closed under the group operation.\n\nTherefore the powers of 7 form a subgroup."}
︠3f6b84a5-d02f-43fd-946c-02e7158ebc67i︠
%md

In the above example there is nothing special about 7 (or even $\mathbb{Z}_n^*$) and so we get the following theorem of group theory:

#### Theorem:  Cyclic subgroup theorem

Let $(G,\circ)$ be a finite group.  Then every element $a \in G$ with $ord(a)=s$ is the generator of a cyclic subgroup with $s$ elements.

*Proof:*

In our subgroup example, replace $\mathbb{Z}_n^*$ with $G$ and $7$ with $a$.

︡4cd9f504-fb62-4951-918d-de68b47e8fc7︡{"done":true,"md":"\nIn the above example there is nothing special about 7 (or even $\\mathbb{Z}_n^*$) and so we get the following theorem of group theory:\n\n#### Theorem:  Cyclic subgroup theorem\n\nLet $(G,\\circ)$ be a finite group.  Then every element $a \\in G$ with $ord(a)=s$ is the generator of a cyclic subgroup with $s$ elements.\n\n*Proof:*\n\nIn our subgroup example, replace $\\mathbb{Z}_n^*$ with $G$ and $7$ with $a$."}
︠36e73712-5039-4bcb-8efe-f86090822cf9i︠
%md

### Enumerating all cyclic subgroups

It's possible for two different elements of $G$ to generate the same cyclic subgroup.



Here is an example of that happening.
︡98be5296-c76a-4fd5-b9a8-e0ca7a70b4f4︡{"done":true,"md":"\n### Enumerating all cyclic subgroups\n\nIt's possible for two different elements of $G$ to generate the same cyclic subgroup.\n\n\n\nHere is an example of that happening."}
︠48621529-f609-469a-8ac9-7a6614060eef︠
## Enumerate all cyclic subgroups of Z_n^*

n = 16
G = Zmod(n)
print("Working modulo {}".format(n))
for a in range(n):
    if gcd(a,n) != 1:
        continue
    print("The subgroup generated by {} is:".format(a))
    subgroup = [power_mod(a,k,n) for k in range(G(a).multiplicative_order())]
    print(sorted(subgroup))
︡3c10a029-cc13-4b45-986c-e777a3f3c691︡{"stdout":"Working modulo 16\n"}︡{"stdout":"The subgroup generated by 1 is:\n[1]\nThe subgroup generated by 3 is:\n[1, 3, 9, 11]\nThe subgroup generated by 5 is:\n[1, 5, 9, 13]\nThe subgroup generated by 7 is:\n[1, 7]\nThe subgroup generated by 9 is:\n[1, 9]\nThe subgroup generated by 11 is:\n[1, 3, 9, 11]\nThe subgroup generated by 13 is:\n[1, 5, 9, 13]\nThe subgroup generated by 15 is:\n[1, 15]\n"}︡{"done":true}
︠2763f670-750d-4057-abdb-456829b58791i︠
%md

### Observation

Notice that the group $\mathbb{Z}_{16}^*$ is not cyclic.

Also observe that some distinct subgroups have the same cardinality.

An example of this is $H=\{1,9\}$ and $K=\{1,15\}$.

We will see below that in a cyclic group this never happens:  No two distinct subgroups have the same cardinality in a cyclic group.

︡2774c4e1-fe28-4bf1-adbb-c5a342edf3fc︡{"done":true,"md":"\n### Observation\n\nNotice that the group $\\mathbb{Z}_{16}^*$ is not cyclic.\n\nAlso observe that some distinct subgroups have the same cardinality.\n\nAn example of this is $H=\\{1,9\\}$ and $K=\\{1,15\\}$.\n\nWe will see below that in a cyclic group this never happens:  No two distinct subgroups have the same cardinality in a cyclic group."}
︠f893f457-1555-4a5d-b1ff-bdf9d8e7d5afi︠
%md

### Studying cyclic subgroups

In the above example where $n=16$, you can see that 3 and 11 generate the same subgroup.

This happens because 3 is itself a primitive element of the subgroup generated by 11 (and vice versa).

Observe that the intersection of two subgroups is always a subgroup!

$\mathbb{Z}_{16}^*$ is not cyclic and so no subgroup is equal to the whole group.

But the whole group is always a subgroup of itself.

Let's try the same experiment with a cyclic group.

︡4d9afe8e-9e98-42b2-8c3b-0ff0d14448a3︡{"done":true,"md":"\n### Studying cyclic subgroups\n\nIn the above example where $n=16$, you can see that 3 and 11 generate the same subgroup.\n\nThis happens because 3 is itself a primitive element of the subgroup generated by 11 (and vice versa).\n\nObserve that the intersection of two subgroups is always a subgroup!\n\n$\\mathbb{Z}_{16}^*$ is not cyclic and so no subgroup is equal to the whole group.\n\nBut the whole group is always a subgroup of itself.\n\nLet's try the same experiment with a cyclic group."}
︠52b534b1-f545-475a-9cb7-63507fe6b6d1︠
## Enumerate all cyclic subgroups of Z_n^*

n = 23
G = Zmod(n)
print("Working modulo {}".format(n))
for a in range(n):
    if gcd(a,n) != 1:
        continue
    print("The subgroup generated by {} is:".format(a))
    subgroup = [power_mod(a,k,n) for k in range(G(a).multiplicative_order())]
    print(sorted(subgroup))
︡a19a15cc-0894-470d-a2db-a19707bb622a︡{"stdout":"Working modulo 23\n"}︡{"stdout":"The subgroup generated by 1 is:\n[1]\nThe subgroup generated by 2 is:\n[1, 2, 3, 4, 6, 8, 9, 12, 13, 16, 18]\nThe subgroup generated by 3 is:\n[1, 2, 3, 4, 6, 8, 9, 12, 13, 16, 18]\nThe subgroup generated by 4 is:\n[1, 2, 3, 4, 6, 8, 9, 12, 13, 16, 18]\nThe subgroup generated by 5 is:\n[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22]\nThe subgroup generated by 6 is:\n[1, 2, 3, 4, 6, 8, 9, 12, 13, 16, 18]\nThe subgroup generated by 7 is:\n[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22]\nThe subgroup generated by 8 is:\n[1, 2, 3, 4, 6, 8, 9, 12, 13, 16, 18]\nThe subgroup generated by 9 is:\n[1, 2, 3, 4, 6, 8, 9, 12, 13, 16, 18]\nThe subgroup generated by 10 is:\n[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22]\nThe subgroup generated by 11 is:\n[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22]\nThe subgroup generated by 12 is:\n[1, 2, 3, 4, 6, 8, 9, 12, 13, 16, 18]\nThe subgroup generated by 13 is:\n[1, 2, 3, 4, 6, 8, 9, 12, 13, 16, 18]\nThe subgroup generated by 14 is:\n[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22]\nThe subgroup generated by 15 is:\n[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22]\nThe subgroup generated by 16 is:\n[1, 2, 3, 4, 6, 8, 9, 12, 13, 16, 18]\nThe subgroup generated by 17 is:\n[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22]\nThe subgroup generated by 18 is:\n[1, 2, 3, 4, 6, 8, 9, 12, 13, 16, 18]\nThe subgroup generated by 19 is:\n[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22]\nThe subgroup generated by 20 is:\n[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22]\nThe subgroup generated by 21 is:\n[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22]\nThe subgroup generated by 22 is:\n[1, 22]\n"}︡{"done":true}
︠d0a190b5-bd1b-4f61-bcff-314355449260i︠
%md

### Thinking about subgroups

In the above example we saw that most subgroups of $\mathbb{Z}_{23}^*$ are big.

That is because the orders of the elements determine the orders of the subgroups.

So any subgroup must be a size that divides $23-1 = 22$.

But $23$ is a safe prime because $22=2\cdot11$  ($p-1$ is twice a prime).

The point of a safe prime is exactly that $\mathbb{Z}_p^*$ only has big subgroups (except for the unique subgroup of size 2, $\{1,-1\}$).

Let's try the same experiment with an "unsafe" prime.

︡21e80281-f4fc-4357-9644-476ec870dd40︡{"done":true,"md":"\n### Thinking about subgroups\n\nIn the above example we saw that most subgroups of $\\mathbb{Z}_{23}^*$ are big.\n\nThat is because the orders of the elements determine the orders of the subgroups.\n\nSo any subgroup must be a size that divides $23-1 = 22$.\n\nBut $23$ is a safe prime because $22=2\\cdot11$  ($p-1$ is twice a prime).\n\nThe point of a safe prime is exactly that $\\mathbb{Z}_p^*$ only has big subgroups (except for the unique subgroup of size 2, $\\{1,-1\\}$).\n\nLet's try the same experiment with an \"unsafe\" prime."}
︠f077f814-7d6a-419b-babb-f865e8c90530︠
## Enumerate all cyclic subgroups of Z_n^*

n = 37
G = Zmod(n)
print("Working modulo {}".format(n))
for a in range(n):
    if gcd(a,n) != 1:
        continue
    print("The subgroup generated by {} is:".format(a))
    subgroup = [power_mod(a,k,n) for k in range(G(a).multiplicative_order())]
    print(sorted(subgroup))
︡81756910-ae44-4281-85d3-08b0f7a7130d︡{"stdout":"Working modulo 37\n"}︡{"stdout":"The subgroup generated by 1 is:\n[1]\nThe subgroup generated by 2 is:\n[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36]\nThe subgroup generated by 3 is:\n[1, 3, 4, 7, 9, 10, 11, 12, 16, 21, 25, 26, 27, 28, 30, 33, 34, 36]\nThe subgroup generated by 4 is:\n[1, 3, 4, 7, 9, 10, 11, 12, 16, 21, 25, 26, 27, 28, 30, 33, 34, 36]\nThe subgroup generated by 5 is:\n[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36]\nThe subgroup generated by 6 is:\n[1, 6, 31, 36]\nThe subgroup generated by 7 is:\n[1, 7, 9, 10, 12, 16, 26, 33, 34]\nThe subgroup generated by 8 is:\n[1, 6, 8, 10, 11, 14, 23, 26, 27, 29, 31, 36]\nThe subgroup generated by 9 is:\n[1, 7, 9, 10, 12, 16, 26, 33, 34]\nThe subgroup generated by 10 is:\n[1, 10, 26]\nThe subgroup generated by 11 is:\n[1, 10, 11, 26, 27, 36]\nThe subgroup generated by 12 is:\n[1, 7, 9, 10, 12, 16, 26, 33, 34]\nThe subgroup generated by 13 is:\n[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36]\nThe subgroup generated by 14 is:\n[1, 6, 8, 10, 11, 14, 23, 26, 27, 29, 31, 36]\nThe subgroup generated by 15 is:\n[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36]\nThe subgroup generated by 16 is:\n[1, 7, 9, 10, 12, 16, 26, 33, 34]\nThe subgroup generated by 17 is:\n[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36]\nThe subgroup generated by 18 is:\n[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36]\nThe subgroup generated by 19 is:\n[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36]\nThe subgroup generated by 20 is:\n[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36]\nThe subgroup generated by 21 is:\n[1, 3, 4, 7, 9, 10, 11, 12, 16, 21, 25, 26, 27, 28, 30, 33, 34, 36]\nThe subgroup generated by 22 is:\n[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36]\nThe subgroup generated by 23 is:\n[1, 6, 8, 10, 11, 14, 23, 26, 27, 29, 31, 36]\nThe subgroup generated by 24 is:\n[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36]\nThe subgroup generated by 25 is:\n[1, 3, 4, 7, 9, 10, 11, 12, 16, 21, 25, 26, 27, 28, 30, 33, 34, 36]\nThe subgroup generated by 26 is:\n[1, 10, 26]\nThe subgroup generated by 27 is:\n[1, 10, 11, 26, 27, 36]\nThe subgroup generated by 28 is:\n[1, 3, 4, 7, 9, 10, 11, 12, 16, 21, 25, 26, 27, 28, 30, 33, 34, 36]\nThe subgroup generated by 29 is:\n[1, 6, 8, 10, 11, 14, 23, 26, 27, 29, 31, 36]\nThe subgroup generated by 30 is:\n[1, 3, 4, 7, 9, 10, 11, 12, 16, 21, 25, 26, 27, 28, 30, 33, 34, 36]\nThe subgroup generated by 31 is:\n[1, 6, 31, 36]\nThe subgroup generated by 32 is:\n[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36]\nThe subgroup generated by 33 is:\n[1, 7, 9, 10, 12, 16, 26, 33, 34]\nThe subgroup generated by 34 is:\n[1, 7, 9, 10, 12, 16, 26, 33, 34]\nThe subgroup generated by 35 is:\n[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36]\nThe subgroup generated by 36 is:\n[1, 36]\n"}︡{"done":true}
︠2a09f85b-c6fd-436c-8f8e-48fa1554f2bci︠
%md

### Thinking about subgroups

In the case of $p=37$, $p-1=36$ has lots of divisors.

For that reason it has many subgroups, and some of them are too small to be safe from brute force on DLP.
︡043f2f1b-e363-4ee2-bbf8-d87bc359bb09︡{"done":true,"md":"\n### Thinking about subgroups\n\nIn the case of $p=37$, $p-1=36$ has lots of divisors. \n\nFor that reason it has many subgroups, and some of them are too small to be safe from brute force on DLP."}
︠71a434ad-bea9-4a3d-851a-5ff56bf5a66ei︠
%md

The following is an interesting group theoretic fact that is true even when $G$ and $H$ are not cyclic:

### Theorem (Lagrange's Theorem)

Let $H$ be a subgroup of $G$.  Then $|H|$ divides $|G|$.

*Proof sketch*

Sets of the form $gH = \{gh: h \in H\}$ partition $G$.
︡757443d8-beba-4e69-96a6-e97ef7a9a312︡{"done":true,"md":"\nThe following is an interesting group theoretic fact that is true even when $G$ and $H$ are not cyclic:\n\n### Theorem (Lagrange's Theorem)\n\nLet $H$ be a subgroup of $G$.  Then $|H|$ divides $|G|$.\n\n*Proof sketch*\n\nSets of the form $gH = \\{gh: h \\in H\\}$ partition $G$."}
︠f637f2b3-c25c-47ed-8990-951e6ac004a3i︠
%md

### Cyclic Subgroup Theorem:

Let $G$ be a finite cyclic group of order $n$ and let $\alpha$ be a generator of $G$.

Then for every integer $k$ that divides $n$ there exists exactly one cyclic subgroup $H$ of $G$ of order $k$.

This subgroups is generated by $\alpha^{n/k}$.

$H$ consists exactly of the elements $a \in G$ that satisfy $a^k = 1$.

There are no other subgroups.



︡ddeb92fd-651b-47ba-ae39-47585068af90︡{"done":true,"md":"\n### Cyclic Subgroup Theorem:\n\nLet $G$ be a finite cyclic group of order $n$ and let $\\alpha$ be a generator of $G$.\n\nThen for every integer $k$ that divides $n$ there exists exactly one cyclic subgroup $H$ of $G$ of order $k$.\n\nThis subgroups is generated by $\\alpha^{n/k}$.  \n\n$H$ consists exactly of the elements $a \\in G$ that satisfy $a^k = 1$. \n\nThere are no other subgroups."}
︠d13e871b-2e49-4600-b22b-cfcb68ed9fda︠
## Sage example

p = 101
G = Zmod(p)
alpha = G.multiplicative_generator()
alpha
︡c84935f4-5f87-4559-a44f-cb8b427128f4︡{"stdout":"2\n"}︡{"done":true}
︠32c72568-3f4d-42a9-98ed-95b4806f1baa︠
divisors(p-1)
︡d99306de-104a-46a3-a957-7cd13088db5b︡{"stdout":"[1, 2, 4, 5, 10, 20, 25, 50, 100]\n"}︡{"done":true}
︠8d084d34-0e9b-45ac-b8b1-78a3d3ef3400︠
for k in divisors(p-1):
    generator = alpha^((p-1)//k)
    subgroup = [generator^j for j in range(generator.multiplicative_order())]
    print("The following subgroup has generator = {} = alpha^{}  (n={} and k={})".format(generator,(p-1)//k,p-1,k))
    print(sorted(subgroup))
︡3ff8a253-9117-46c5-8596-243f6edea089︡{"stdout":"The following subgroup has generator = 1 = alpha^100  (n=100 and k=1)\n[1]\nThe following subgroup has generator = 100 = alpha^50  (n=100 and k=2)\n[1, 100]\nThe following subgroup has generator = 10 = alpha^25  (n=100 and k=4)\n[1, 10, 91, 100]\nThe following subgroup has generator = 95 = alpha^20  (n=100 and k=5)\n[1, 36, 84, 87, 95]\nThe following subgroup has generator = 14 = alpha^10  (n=100 and k=10)\n[1, 6, 14, 17, 36, 65, 84, 87, 95, 100]\nThe following subgroup has generator = 32 = alpha^5  (n=100 and k=20)\n[1, 6, 10, 14, 17, 32, 36, 39, 41, 44, 57, 60, 62, 65, 69, 84, 87, 91, 95, 100]\nThe following subgroup has generator = 16 = alpha^4  (n=100 and k=25)\n[1, 5, 16, 19, 24, 25, 31, 36, 37, 52, 54, 56, 58, 68, 71, 78, 79, 80, 81, 84, 87, 88, 92, 95, 97]\nThe following subgroup has generator = 4 = alpha^2  (n=100 and k=50)\n[1, 4, 5, 6, 9, 13, 14, 16, 17, 19, 20, 21, 22, 23, 24, 25, 30, 31, 33, 36, 37, 43, 45, 47, 49, 52, 54, 56, 58, 64, 65, 68, 70, 71, 76, 77, 78, 79, 80, 81, 82, 84, 85, 87, 88, 92, 95, 96, 97, 100]\nThe following subgroup has generator = 2 = alpha^1  (n=100 and k=100)\n[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100]\n"}︡{"done":true}
︠b4ba06a5-19b3-4e38-9971-b6f04c78ba12i︠
%md

###  Safe primes again

Okay, so what are the subgroups of a safe prime?

Recall that $p$ is safe if $p-1 = 2q$ where $q$ is prime.

You can see from the definition that the only divisors of $p-1$ are 1, 2, $q$, and $p-1$.

Let's see an example.

︡a404190a-7960-470f-80cf-66a3b3ff683a︡{"done":true,"md":"\n###  Safe primes again\n\nOkay, so what are the subgroups of a safe prime?\n\nRecall that $p$ is safe if $p-1 = 2q$ where $q$ is prime.\n\nYou can see from the definition that the only divisors of $p-1$ are 1, 2, $q$, and $p-1$.\n\nLet's see an example."}
︠a3c490da-0f63-4fa9-a229-83bf627f6817s︠

def next_safe_prime(n):
    n = next_prime(n)
    while True:
        if is_prime((n-1)//2):
            break
        n = next_prime(n+1)
        print(n,factor(n-1))
    return n

next_safe_prime(100)
︡1323cb28-c03f-420e-ab1d-acb08bf66a60︡{"stdout":"(103, 2 * 3 * 17)\n(107, 2 * 53)\n107\n"}︡{"done":true}
︠c17a44f7-c72a-462b-945e-a812ced635bds︠
p = 107
G = Zmod(p)
alpha = G.multiplicative_generator()
alpha
︡33c24dd3-e9af-48d6-a857-81ba1063b15f︡{"stdout":"2\n"}︡{"done":true}
︠4f9eefe5-2877-41ea-8852-558be62ca34cs︠
divisors(p-1)
︡4e43f9d0-ca89-4ed4-9b6f-e1c1bfa89800︡{"stdout":"[1, 2, 53, 106]\n"}︡{"done":true}
︠5e97c60c-b2b4-4f5d-87f8-02d7b1f34be8s︠
for k in divisors(p-1):
    generator = alpha^((p-1)//k)
    subgroup =   [generator^j for j in range(generator.multiplicative_order())]
    print("The following subgroup has generator = {} = alpha^{}  (n={} and k={})".format(generator,(p-1)//k,p-1,k))
    print(sorted(subgroup))
︡6fb4dd9b-0a35-4a52-8f4b-8d471d6c7cf0︡{"stdout":"The following subgroup has generator = 1 = alpha^106  (n=106 and k=1)\n[1]\nThe following subgroup has generator = 106 = alpha^53  (n=106 and k=2)\n[1, 106]\nThe following subgroup has generator = 4 = alpha^2  (n=106 and k=53)\n[1, 3, 4, 9, 10, 11, 12, 13, 14, 16, 19, 23, 25, 27, 29, 30, 33, 34, 35, 36, 37, 39, 40, 41, 42, 44, 47, 48, 49, 52, 53, 56, 57, 61, 62, 64, 69, 75, 76, 79, 81, 83, 85, 86, 87, 89, 90, 92, 99, 100, 101, 102, 105]\nThe following subgroup has generator = 2 = alpha^1  (n=106 and k=106)\n[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 104, 105, 106]\n"}︡{"done":true}
︠3d5a75e6-ecc1-412f-bfa8-c83c27e34fcdi︠
%md

### Safe primes

You can see from the above output that the only really dangerous base for the DLP in this group is the element of order 2.

The number of bits in $q$ is only one less than $p$, so even if the base is in the subgroup of size $q$, this is okay.

And it's really easy to check whether $\alpha$ has order 2:

Just see if $\alpha^2 = 1$.

And since $p$ is prime this is only true for $\alpha = p-1$.

It's similarly easy to see if the order of $\alpha$ is $q$.

Thus for a safe prime we can find a verifiable generator.

︡fa815638-7939-4b90-9843-d34832ce3a3d︡{"done":true,"md":"\n### Safe primes\n\nYou can see from the above output that the only really dangerous base for the DLP in this group is the element of order 2.\n\nThe number of bits in $q$ is only one less than $p$, so even if the base is in the subgroup of size $q$, this is okay.\n\nAnd it's really easy to check whether $\\alpha$ has order 2: \n\nJust see if $\\alpha^2 = 1$. \n\nAnd since $p$ is prime this is only true for $\\alpha = p-1$. \n\nIt's similarly easy to see if the order of $\\alpha$ is $q$.\n\nThus for a safe prime we can find a verifiable generator."}
︠d4d40d4b-871f-4053-b233-dfc35e2c983ei︠
%md

### Diffie Helman in openssl

Below we will verify that the Diffie Hellman parameters produced by OpenSSL yield a safe prime.

Because of this fact generating DH params takes much longer than RSA params, even if the bit length is the same.


︡831b5d3d-eb6b-48c7-916c-c922ab9397c9︡{"done":true,"md":"\n### Diffie Helman in openssl\n\nBelow we will verify that the Diffie Hellman parameters produced by OpenSSL yield a safe prime.\n\nBecause of this fact generating DH params takes much longer than RSA params, even if the bit length is the same."}
︠24c15e41-a8bb-4c01-ab6d-33f4c6c68aa8i︠
%md

```
hunter@pop-os:~/.ssh$ openssl dhparam -out dh.pem 512
Generating DH parameters, 512 bit long safe prime, generator 2
This is going to take a long time
......................................................+......................+........++*++*++*++*++*
hunter@pop-os:~/.ssh$ openssl dhparam -text -in dh.pem
    DH Parameters: (512 bit)
        prime:
            00:e2:e8:54:d3:75:f1:0a:2d:90:c1:5d:54:82:1e:
            b5:75:f7:23:06:41:7c:d3:d9:23:9c:39:b7:ae:fb:
            a0:eb:9d:be:93:da:65:0a:ea:13:84:49:eb:04:11:
            ad:dc:dc:62:41:2a:aa:92:bb:52:02:a0:37:14:0b:
            cc:fa:2d:36:43
        generator: 2 (0x2)
-----BEGIN DH PARAMETERS-----
MEYCQQDi6FTTdfEKLZDBXVSCHrV19yMGQXzT2SOcObeu+6Drnb6T2mUK6hOESesE
Ea3c3GJBKqqSu1ICoDcUC8z6LTZDAgEC
-----END DH PARAMETERS-----
```
︡bdfb4768-e36e-4149-9d8b-4a59d0aa59fb︡{"done":true,"md":"\n```\nhunter@pop-os:~/.ssh$ openssl dhparam -out dh.pem 512\nGenerating DH parameters, 512 bit long safe prime, generator 2\nThis is going to take a long time\n......................................................+......................+........++*++*++*++*++*\nhunter@pop-os:~/.ssh$ openssl dhparam -text -in dh.pem\n    DH Parameters: (512 bit)\n        prime:\n            00:e2:e8:54:d3:75:f1:0a:2d:90:c1:5d:54:82:1e:\n            b5:75:f7:23:06:41:7c:d3:d9:23:9c:39:b7:ae:fb:\n            a0:eb:9d:be:93:da:65:0a:ea:13:84:49:eb:04:11:\n            ad:dc:dc:62:41:2a:aa:92:bb:52:02:a0:37:14:0b:\n            cc:fa:2d:36:43\n        generator: 2 (0x2)\n-----BEGIN DH PARAMETERS-----\nMEYCQQDi6FTTdfEKLZDBXVSCHrV19yMGQXzT2SOcObeu+6Drnb6T2mUK6hOESesE\nEa3c3GJBKqqSu1ICoDcUC8z6LTZDAgEC\n-----END DH PARAMETERS-----\n```"}
︠a3d5ec4b-ba26-4b5e-ad40-32ca1dec30d9s︠

︡2787372b-77ca-4a07-a1c9-e4c7baa694c3︡{"done":true}
︠d2e582a8-54d9-4a7e-9973-c33476acefe6s︠

p = """00:e2:e8:54:d3:75:f1:0a:2d:90:c1:5d:54:82:1e:
            b5:75:f7:23:06:41:7c:d3:d9:23:9c:39:b7:ae:fb:
            a0:eb:9d:be:93:da:65:0a:ea:13:84:49:eb:04:11:
            ad:dc:dc:62:41:2a:aa:92:bb:52:02:a0:37:14:0b:
            cc:fa:2d:36:43"""
p = p.replace(':','').replace('\n','').replace('\t','').replace(' ','')
p = int(p,16)
p
︡00cdb757-8da0-4597-8923-43515989a0e8︡{"stdout":"11884112392174931525551803878284316468547743675014443360835956479033078578449820213217260145091693859837332463855747972651088419742292198211998500015388227L\n"}︡{"done":true}
︠e60dd8d3-ed3f-4fe2-9be3-525bf531bfdcs︠
is_prime(p)
︡101236f4-5c75-4b62-a5eb-dd19010735d3︡{"stdout":"True"}︡{"stdout":"\n"}︡{"done":true}
︠a34cbd91-621b-4e6e-8df2-f4d3aa596e2cs︠
q = (p-1)//2
q
︡6155b411-a674-47e3-9597-ce0e55cf88ef︡{"stdout":"5942056196087465762775901939142158234273871837507221680417978239516539289224910106608630072545846929918666231927873986325544209871146099105999250007694113\n"}︡{"done":true}
︠9af916f6-4363-4b7e-bf36-36907e8a2127s︠
is_prime(q)
︡433c8e8a-2b20-48b3-b36a-38922e1bdf4c︡{"stdout":"True"}︡{"stdout":"\n"}︡{"done":true}
︠5927baf8-741d-4c8e-bfff-4d5333c2377as︠
G=Zmod(p)
alpha = G(2)
alpha^(q)
︡45ea5f06-366a-4aa1-bd4d-b65f21ef60e7︡{"stdout":"11884112392174931525551803878284316468547743675014443360835956479033078578449820213217260145091693859837332463855747972651088419742292198211998500015388226\n"}︡{"done":true}
︠edd1f6e7-2c7e-486b-b64d-fefe5c21c68di︠
%md

### Safe prime

In the above experiment $\alpha^q = p-1$.

Thus we can be sure that the order of $\alpha$ is $p$, because if it were in the subgroup of size $q$ its order would have to divide $q$.

But in that case we would have observed $\alpha^q =1$.

︡87b86ed5-61f4-4cd8-b08f-93af26e001f2︡{"done":true,"md":"\n### Safe prime\n\nIn the above experiment $\\alpha^q = p-1$.  \n\nThus we can be sure that the order of $\\alpha$ is $p$, because if it were in the subgroup of size $q$ its order would have to divide $q$.\n\nBut in that case we would have observed $\\alpha^q =1$."}
︠c85458d9-277f-4f6d-8abc-d33d00a77f7di︠
%md

### The general DLP

The DLP is a problem that can be stated in any group (not just our simple "mod p") groups.

### Definition:  The Generalized Discrete Logarithm Problem

Given is a finite cyclic group $G$ with the group operation $\circ$ and cardinality $n$.

We consider a primitive element $\alpha \in G$ and another element $\beta \in G$.

The DLP is finding the integer $x$ where $1 \leq x \leq n$ such that :

$$\beta = \underbrace{\alpha\circ \alpha \circ \cdots \circ \alpha}_{x\, times} = \alpha^x$$
︡074e8e3f-1c11-4028-8386-88c91f3509c0︡{"done":true,"md":"\n### The general DLP\n\nThe DLP is a problem that can be stated in any group (not just our simple \"mod p\") groups.\n\n### Definition:  The Generalized Discrete Logarithm Problem\n\nGiven is a finite cyclic group $G$ with the group operation $\\circ$ and cardinality $n$. \n\nWe consider a primitive element $\\alpha \\in G$ and another element $\\beta \\in G$.\n\nThe DLP is finding the integer $x$ where $1 \\leq x \\leq n$ such that :\n\n$$\\beta = \\underbrace{\\alpha\\circ \\alpha \\circ \\cdots \\circ \\alpha}_{x\\, times} = \\alpha^x$$"}
︠e3e4ff4c-abfb-451e-9d45-2dda43cd4c53i︠
%md

### Elliptic Curve groups

In particular you can have the DLP in an elliptic curve group.

We won't give the details for this.

But:  If we use DLP in $\mathbb{Z}_p^*$ and also DLP in elliptic curve groups, why are the keys for elliptic curve groups so much shorter?

We've seen that for elliptic curves, 256 bits gives 128 bit security, while for $\mathbb{Z}_p^*$ we need about 3000 bits.

Why is one group vulnerable to DLP attacks and the other group not?


︡d14efbf2-f185-4a32-9d46-a8322509409f︡{"done":true,"md":"\n### Elliptic Curve groups\n\nIn particular you can have the DLP in an elliptic curve group.  \n\nWe won't give the details for this.\n\nBut:  If we use DLP in $\\mathbb{Z}_p^*$ and also DLP in elliptic curve groups, why are the keys for elliptic curve groups so much shorter?\n\nWe've seen that for elliptic curves, 256 bits gives 128 bit security, while for $\\mathbb{Z}_p^*$ we need about 3000 bits.  \n\nWhy is one group vulnerable to DLP attacks and the other group not?"}
︠9c00b32e-9734-4d49-8784-687a917fa263i︠
%md

### Index calculus

The main reason key sizes for elliptic curve groups are smaller is that $\mathbb{Z}_p^*$ is vulnerable to an attack called the [index calculus](https://en.wikipedia.org/wiki/Index_calculus_algorithm).

Again, we won't go into the details, but you can study this on your own if you want.

No one knows of any way to apply this attack to the DLP in elliptic curve groups.


There is some speculation however that the parameters for elliptic curve groups are backdoored by NSA.

However if this is true, they have not yet bothered to destroy any of the dozens of cryptocurrencies based on a NIST curve.

︡0b222133-a215-4ce8-9b19-6bc4d74ff4d5︡{"done":true,"md":"\n### Index calculus\n\nThe main reason key sizes for elliptic curve groups are smaller is that $\\mathbb{Z}_p^*$ is vulnerable to an attack called the [index calculus](https://en.wikipedia.org/wiki/Index_calculus_algorithm).\n\nAgain, we won't go into the details, but you can study this on your own if you want. \n\nNo one knows of any way to apply this attack to the DLP in elliptic curve groups.\n\n\nThere is some speculation however that the parameters for elliptic curve groups are backdoored by NSA.\n\nHowever if this is true, they have not yet bothered to destroy any of the dozens of cryptocurrencies based on a NIST curve."}
︠3f2fe855-9515-4818-ba31-3270ae3356fds︠
### Below we use Sage to investigate an elliptic curve E with 10 elements.

### We show that (0:7:1) is a generator.
### Note that elliptic curve groups use additive notation.

︡e8906c2b-e03e-4e58-864a-7f0214c90d7a︡{"done":true}
︠2a03193e-ea65-493f-86e6-5579bb2c6de8s︠
E=EllipticCurve(GF(11),[2,5])
E
︡ca4b034b-1d96-4429-a805-4d2d5767b243︡{"stdout":"Elliptic Curve defined by y^2 = x^3 + 2*x + 5 over Finite Field of size 11\n"}︡{"done":true}
︠508602d7-c48b-45bc-8dfe-29ddf111b5a1s︠
E.cardinality()
︡cbe52ccb-ba2c-4f6b-ad25-e884634da38b︡{"stdout":"10\n"}︡{"done":true}
︠3d60d53d-0807-45c7-95bb-869d48cd8b7bs︠
g = E.an_element()
g
︡6c68f52b-cc73-4935-87d9-93de51356853︡{"stdout":"(0 : 7 : 1)\n"}︡{"done":true}
︠3c996db1-f871-43a7-b349-e5ff1b1024a8s︠
2*g
︡e4266b17-db05-4e73-9ed9-dd2edac8a269︡{"stdout":"(9 : 9 : 1)\n"}︡{"done":true}
︠8863eef2-596b-4b9b-b713-4df2e7fea1f7s︠
3*g
︡ec0584f6-d4d3-4cab-8ff9-2826bc0a9401︡{"stdout":"(3 : 7 : 1)\n"}︡{"done":true}
︠0f42a9c8-cc43-4b21-8c9a-5519f310591es︠
4*g
︡db8722c8-efc2-474b-b9a7-ebf00a81c164︡{"stdout":"(8 : 4 : 1)\n"}︡{"done":true}
︠95cdcfbe-7349-4f4a-9275-00bd0d217205s︠
5*g
︡c6e34196-50e9-438e-ba11-d4d5f471eb9a︡{"stdout":"(4 : 0 : 1)\n"}︡{"done":true}
︠4e498e34-692c-4290-9a0a-12ac8f9ff3bas︠
6*g
︡5e11c746-cb5a-4e98-8b2d-eab2626a99f2︡{"stdout":"(8 : 7 : 1)\n"}︡{"done":true}
︠f771a70e-5ccb-4eaa-afc7-c2f3e90f46acs︠
7*g
︡88f6b104-c4b8-4360-8f32-0f94f2eb28f4︡{"stdout":"(3 : 4 : 1)\n"}︡{"done":true}
︠0cd0dcb2-b6de-48df-ab1c-fdcf50ebf402s︠
8*g
︡a6ef3d7e-afbc-4bb6-aece-2df9adf1aa73︡{"stdout":"(9 : 2 : 1)\n"}︡{"done":true}
︠ac9896cb-dbbf-4448-b8cb-3784d7834ef3s︠
9*g
︡bee96eef-3d42-4314-8fa0-cb3f08940f96︡{"stdout":"(0 : 4 : 1)\n"}︡{"done":true}
︠e683dde0-dc4f-42b4-a593-e51aa3ddd053s︠
10*g
︡ee58604c-49c7-474b-acc5-6e6460ea558d︡{"stdout":"(0 : 1 : 0)\n"}︡{"done":true}
︠5977d0cb-31d6-4a70-886a-87667e83276fi︠
%md

### Cyclic group theorem applied to elliptic curves...

The Cyclic Group theorem still applies to $E$ even though $E$ is "weird".

That's because the theorem applies to **any** cyclic group, no matter how weird.

︡cb03ba3f-dbd8-42da-9dcc-603dee6e3014︡{"done":true,"md":"\n### Cyclic group theorem applied to elliptic curves...\n\nThe Cyclic Group theorem still applies to $E$ even though $E$ is \"weird\".\n\nThat's because the theorem applies to **any** cyclic group, no matter how weird."}
︠7a3bf1bf-f136-45bf-ae42-93e66ecfd5cc︠
︡2e578210-4592-41d5-9dfc-730dbffdd3d8︡
︠531a3360-7a47-4749-bac3-0ff87afccd53s︠
for g in E:
    print("The order of {} is {}".format(g,g.order()))
︡409626cb-ed71-486f-939f-060ab9cdec93︡{"stdout":"The order of (0 : 1 : 0) is 1\nThe order of (0 : 4 : 1) is 10\nThe order of (0 : 7 : 1) is 10\nThe order of (3 : 4 : 1) is 10\nThe order of (3 : 7 : 1) is 10\nThe order of (4 : 0 : 1) is 2\nThe order of (8 : 4 : 1) is 5\nThe order of (8 : 7 : 1) is 5\nThe order of (9 : 2 : 1) is 5\nThe order of (9 : 9 : 1) is 5\n"}︡{"done":true}
︠5dfe860a-e8dd-4738-b4b8-abf8f6f0f4e7s︠
euler_phi(10)
︡9b3bbb75-fa64-4107-b33a-8e10684eac51︡{"stdout":"4\n"}︡{"done":true}
︠e62f3ab9-3b10-493a-ba59-8d2090b178e2s︠
alpha = E.an_element()
for k in divisors(10):
    generator = (10//k)*alpha
    subgroup =   [j*generator for j in range(generator.order())]
    print("The following subgroup has generator = {} = {}*alpha   (n={} and k={})".format(generator,(10//k),10,k))
    print(sorted(subgroup))
︡a55eaa1c-bcb7-4669-a694-3e5d1466e303︡{"stdout":"The following subgroup has generator = (0 : 1 : 0) = 10*alpha   (n=10 and k=1)\n[(0 : 1 : 0)]\nThe following subgroup has generator = (4 : 0 : 1) = 5*alpha   (n=10 and k=2)\n[(0 : 1 : 0), (4 : 0 : 1)]\nThe following subgroup has generator = (9 : 9 : 1) = 2*alpha   (n=10 and k=5)\n[(0 : 1 : 0), (8 : 4 : 1), (8 : 7 : 1), (9 : 2 : 1), (9 : 9 : 1)]\nThe following subgroup has generator = (0 : 7 : 1) = 1*alpha   (n=10 and k=10)\n[(0 : 1 : 0), (0 : 4 : 1), (0 : 7 : 1), (3 : 4 : 1), (3 : 7 : 1), (4 : 0 : 1), (8 : 4 : 1), (8 : 7 : 1), (9 : 2 : 1), (9 : 9 : 1)]\n"}︡{"done":true}
︠0815008f-80ee-4e90-94c2-6ae37c2c28ffi︠
%md

### The cyclic group theorem applied to $GF(2^8)$

It turns out Galois Fields are cyclic. [paper](http://www.math.wm.edu/~vinroot/430S13MultFiniteField.pdf)
︡08419cb4-7c2b-44d6-bfbd-be60fc648e1d︡{"done":true,"md":"\n### The cyclic group theorem applied to $GF(2^8)$\n\nIt turns out Galois Fields are cyclic. [paper](http://www.math.wm.edu/~vinroot/430S13MultFiniteField.pdf)"}
︠2aebeb10-b8f0-4fe4-8edc-f4eee7b22566︠
n = 256

G = GF(n,'x')
alpha = [f for f in G if f != 0 and f.multiplicative_order()==n-1][0]  ## Get a generator
print("alpha = {}, alpha.order() = {}".format(alpha,alpha.multiplicative_order()))
for k in divisors((n-1)):
    generator = alpha^((n-1)//k)
    subgroup =   [generator^j for j in range(generator.multiplicative_order())]
    print("The following subgroup has generator = {} = alpha^{}   (|G|={} and k={})".format(generator,((n-1)//k),(n-1),k))
    print(sorted(subgroup))
︡c23cf13b-e78b-4d4c-8943-17ee8637db6a︡{"stdout":"alpha = x, alpha.order() = 255\n"}︡{"stdout":"The following subgroup has generator = 1 = alpha^255   (|G|=255 and k=1)\n[1]\nThe following subgroup has generator = x^7 + x^6 + x^4 + x^2 + x = alpha^85   (|G|=255 and k=3)\n[1, x^7 + x^6 + x^4 + x^2 + x, x^7 + x^6 + x^4 + x^2 + x + 1]\nThe following subgroup has generator = x^3 + x = alpha^51   (|G|=255 and k=5)\n[1, x^3 + x, x^6 + x^2, x^7 + x^4 + x, x^7 + x^6 + x^4 + x^3 + x^2 + 1]\nThe following subgroup has generator = x^7 + x^4 + x^3 = alpha^17   (|G|=255 and k=15)\n[1, x^3 + x, x^3 + x + 1, x^6 + x^2, x^6 + x^2 + 1, x^6 + x^3 + x^2 + x, x^6 + x^3 + x^2 + x + 1, x^7 + x^4 + x, x^7 + x^4 + x + 1, x^7 + x^4 + x^3, x^7 + x^4 + x^3 + 1, x^7 + x^6 + x^4 + x^2 + x, x^7 + x^6 + x^4 + x^2 + x + 1, x^7 + x^6 + x^4 + x^3 + x^2, x^7 + x^6 + x^4 + x^3 + x^2 + 1]\nThe following subgroup has generator = x^5 + x^2 + x = alpha^15   (|G|=255 and k=17)\n[1, x^3 + x^2 + x + 1, x^4 + x^3 + x, x^5 + x^2, x^5 + x^2 + x, x^5 + x^3 + x^2, x^5 + x^4 + x^3 + x + 1, x^6 + x^4 + x^2 + 1, x^6 + x^4 + x^3 + 1, x^6 + x^5, x^6 + x^5 + x^2, x^7 + x^4 + 1, x^7 + x^4 + x^2 + x, x^7 + x^5 + x^3 + 1, x^7 + x^5 + x^4 + x^3 + 1, x^7 + x^6 + 1, x^7 + x^6 + x^4 + x^3 + x^2 + x + 1]\nThe following subgroup has generator = x^5 = alpha^5   (|G|=255 and k=51)\n[1, x + 1, x^2 + 1, x^3 + x^2 + x + 1, x^4 + 1, x^4 + x^3 + x, x^4 + x^3 + x^2, x^5, x^5 + x^2, x^5 + x^2 + x, x^5 + x^3 + x^2, x^5 + x^3 + x^2 + x, x^5 + x^4 + x + 1, x^5 + x^4 + x^2 + x + 1, x^5 + x^4 + x^3 + x + 1, x^6 + x^3 + x^2 + 1, x^6 + x^4 + x^2 + 1, x^6 + x^4 + x^3 + 1, x^6 + x^4 + x^3 + x^2 + x, x^6 + x^5, x^6 + x^5 + x^2, x^6 + x^5 + x^2 + x + 1, x^6 + x^5 + x^3 + x, x^6 + x^5 + x^3 + x^2, x^6 + x^5 + x^4 + x, x^6 + x^5 + x^4 + x^2, x^6 + x^5 + x^4 + x^3 + x^2, x^7 + x^2, x^7 + x^4 + 1, x^7 + x^4 + x^2 + x, x^7 + x^4 + x^3 + x^2, x^7 + x^5, x^7 + x^5 + x^2 + x + 1, x^7 + x^5 + x^3 + 1, x^7 + x^5 + x^3 + x^2, x^7 + x^5 + x^3 + x^2 + x, x^7 + x^5 + x^4 + x^2, x^7 + x^5 + x^4 + x^3 + 1, x^7 + x^5 + x^4 + x^3 + x^2 + x, x^7 + x^6 + 1, x^7 + x^6 + x^4 + x^2 + x, x^7 + x^6 + x^4 + x^2 + x + 1, x^7 + x^6 + x^4 + x^3 + x^2 + x + 1, x^7 + x^6 + x^5 + x, x^7 + x^6 + x^5 + x^2 + x, x^7 + x^6 + x^5 + x^3 + 1, x^7 + x^6 + x^5 + x^3 + x + 1, x^7 + x^6 + x^5 + x^3 + x^2 + x + 1, x^7 + x^6 + x^5 + x^4 + x^2, x^7 + x^6 + x^5 + x^4 + x^3 + x^2 + 1, x^7 + x^6 + x^5 + x^4 + x^3 + x^2 + x + 1]\nThe following subgroup has generator = x^3 = alpha^3   (|G|=255 and k=85)\n[1, x^2 + x + 1, x^3, x^3 + x, x^3 + x^2, x^3 + x^2 + x + 1, x^4 + x^2 + 1, x^4 + x^2 + x + 1, x^4 + x^3 + x, x^5 + 1, x^5 + x^2, x^5 + x^2 + 1, x^5 + x^2 + x, x^5 + x^2 + x + 1, x^5 + x^3 + 1, x^5 + x^3 + x^2, x^5 + x^3 + x^2 + 1, x^5 + x^3 + x^2 + x + 1, x^5 + x^4 + x^2 + 1, x^5 + x^4 + x^2 + x, x^5 + x^4 + x^3, x^5 + x^4 + x^3 + x, x^5 + x^4 + x^3 + x + 1, x^5 + x^4 + x^3 + x^2 + 1, x^5 + x^4 + x^3 + x^2 + x, x^6, x^6 + x^2, x^6 + x^2 + x, x^6 + x^4, x^6 + x^4 + x^2 + 1, x^6 + x^4 + x^2 + x, x^6 + x^4 + x^2 + x + 1, x^6 + x^4 + x^3 + 1, x^6 + x^5, x^6 + x^5 + 1, x^6 + x^5 + x^2, x^6 + x^5 + x^2 + 1, x^6 + x^5 + x^2 + x, x^6 + x^5 + x^3 + x + 1, x^6 + x^5 + x^3 + x^2 + x, x^6 + x^5 + x^4 + x + 1, x^6 + x^5 + x^4 + x^2 + 1, x^6 + x^5 + x^4 + x^3, x^6 + x^5 + x^4 + x^3 + x^2 + 1, x^6 + x^5 + x^4 + x^3 + x^2 + x + 1, x^7 + x, x^7 + x^2 + x, x^7 + x^3 + x, x^7 + x^3 + x + 1, x^7 + x^3 + x^2 + x + 1, x^7 + x^4 + 1, x^7 + x^4 + x, x^7 + x^4 + x^2 + x, x^7 + x^5 + 1, x^7 + x^5 + x^2 + x, x^7 + x^5 + x^3, x^7 + x^5 + x^3 + 1, x^7 + x^5 + x^3 + x^2 + 1, x^7 + x^5 + x^4 + x + 1, x^7 + x^5 + x^4 + x^2 + 1, x^7 + x^5 + x^4 + x^2 + x, x^7 + x^5 + x^4 + x^3, x^7 + x^5 + x^4 + x^3 + 1, x^7 + x^5 + x^4 + x^3 + x, x^7 + x^5 + x^4 + x^3 + x^2 + x + 1, x^7 + x^6 + 1, x^7 + x^6 + x + 1, x^7 + x^6 + x^2, x^7 + x^6 + x^2 + 1, x^7 + x^6 + x^3 + x^2 + 1, x^7 + x^6 + x^3 + x^2 + x, x^7 + x^6 + x^3 + x^2 + x + 1, x^7 + x^6 + x^4, x^7 + x^6 + x^4 + x^3 + 1, x^7 + x^6 + x^4 + x^3 + x + 1, x^7 + x^6 + x^4 + x^3 + x^2 + 1, x^7 + x^6 + x^4 + x^3 + x^2 + x + 1, x^7 + x^6 + x^5 + x^2, x^7 + x^6 + x^5 + x^2 + x + 1, x^7 + x^6 + x^5 + x^3 + x^2 + 1, x^7 + x^6 + x^5 + x^4 + 1, x^7 + x^6 + x^5 + x^4 + x, x^7 + x^6 + x^5 + x^4 + x^2 + 1, x^7 + x^6 + x^5 + x^4 + x^3 + x + 1, x^7 + x^6 + x^5 + x^4 + x^3 + x^2]"}︡{"stdout":"\nThe following subgroup has generator = x = alpha^1   (|G|=255 and k=255)\n[1, x, x + 1, x^2, x^2 + 1, x^2 + x, x^2 + x + 1, x^3, x^3 + 1, x^3 + x, x^3 + x + 1, x^3 + x^2, x^3 + x^2 + 1, x^3 + x^2 + x, x^3 + x^2 + x + 1, x^4, x^4 + 1, x^4 + x, x^4 + x + 1, x^4 + x^2, x^4 + x^2 + 1, x^4 + x^2 + x, x^4 + x^2 + x + 1, x^4 + x^3, x^4 + x^3 + 1, x^4 + x^3 + x, x^4 + x^3 + x + 1, x^4 + x^3 + x^2, x^4 + x^3 + x^2 + 1, x^4 + x^3 + x^2 + x, x^4 + x^3 + x^2 + x + 1, x^5, x^5 + 1, x^5 + x, x^5 + x + 1, x^5 + x^2, x^5 + x^2 + 1, x^5 + x^2 + x, x^5 + x^2 + x + 1, x^5 + x^3, x^5 + x^3 + 1, x^5 + x^3 + x, x^5 + x^3 + x + 1, x^5 + x^3 + x^2, x^5 + x^3 + x^2 + 1, x^5 + x^3 + x^2 + x, x^5 + x^3 + x^2 + x + 1, x^5 + x^4, x^5 + x^4 + 1, x^5 + x^4 + x, x^5 + x^4 + x + 1, x^5 + x^4 + x^2, x^5 + x^4 + x^2 + 1, x^5 + x^4 + x^2 + x, x^5 + x^4 + x^2 + x + 1, x^5 + x^4 + x^3, x^5 + x^4 + x^3 + 1, x^5 + x^4 + x^3 + x, x^5 + x^4 + x^3 + x + 1, x^5 + x^4 + x^3 + x^2, x^5 + x^4 + x^3 + x^2 + 1, x^5 + x^4 + x^3 + x^2 + x, x^5 + x^4 + x^3 + x^2 + x + 1, x^6, x^6 + 1, x^6 + x, x^6 + x + 1, x^6 + x^2, x^6 + x^2 + 1, x^6 + x^2 + x, x^6 + x^2 + x + 1, x^6 + x^3, x^6 + x^3 + 1, x^6 + x^3 + x, x^6 + x^3 + x + 1, x^6 + x^3 + x^2, x^6 + x^3 + x^2 + 1, x^6 + x^3 + x^2 + x, x^6 + x^3 + x^2 + x + 1, x^6 + x^4, x^6 + x^4 + 1, x^6 + x^4 + x, x^6 + x^4 + x + 1, x^6 + x^4 + x^2, x^6 + x^4 + x^2 + 1, x^6 + x^4 + x^2 + x, x^6 + x^4 + x^2 + x + 1, x^6 + x^4 + x^3, x^6 + x^4 + x^3 + 1, x^6 + x^4 + x^3 + x, x^6 + x^4 + x^3 + x + 1, x^6 + x^4 + x^3 + x^2, x^6 + x^4 + x^3 + x^2 + 1, x^6 + x^4 + x^3 + x^2 + x, x^6 + x^4 + x^3 + x^2 + x + 1, x^6 + x^5, x^6 + x^5 + 1, x^6 + x^5 + x, x^6 + x^5 + x + 1, x^6 + x^5 + x^2, x^6 + x^5 + x^2 + 1, x^6 + x^5 + x^2 + x, x^6 + x^5 + x^2 + x + 1, x^6 + x^5 + x^3, x^6 + x^5 + x^3 + 1, x^6 + x^5 + x^3 + x, x^6 + x^5 + x^3 + x + 1, x^6 + x^5 + x^3 + x^2, x^6 + x^5 + x^3 + x^2 + 1, x^6 + x^5 + x^3 + x^2 + x, x^6 + x^5 + x^3 + x^2 + x + 1, x^6 + x^5 + x^4, x^6 + x^5 + x^4 + 1, x^6 + x^5 + x^4 + x, x^6 + x^5 + x^4 + x + 1, x^6 + x^5 + x^4 + x^2, x^6 + x^5 + x^4 + x^2 + 1, x^6 + x^5 + x^4 + x^2 + x, x^6 + x^5 + x^4 + x^2 + x + 1, x^6 + x^5 + x^4 + x^3, x^6 + x^5 + x^4 + x^3 + 1, x^6 + x^5 + x^4 + x^3 + x, x^6 + x^5 + x^4 + x^3 + x + 1, x^6 + x^5 + x^4 + x^3 + x^2, x^6 + x^5 + x^4 + x^3 + x^2 + 1, x^6 + x^5 + x^4 + x^3 + x^2 + x, x^6 + x^5 + x^4 + x^3 + x^2 + x + 1, x^7, x^7 + 1, x^7 + x, x^7 + x + 1, x^7 + x^2, x^7 + x^2 + 1, x^7 + x^2 + x, x^7 + x^2 + x + 1, x^7 + x^3, x^7 + x^3 + 1, x^7 + x^3 + x, x^7 + x^3 + x + 1, x^7 + x^3 + x^2, x^7 + x^3 + x^2 + 1, x^7 + x^3 + x^2 + x, x^7 + x^3 + x^2 + x + 1, x^7 + x^4, x^7 + x^4 + 1, x^7 + x^4 + x, x^7 + x^4 + x + 1, x^7 + x^4 + x^2, x^7 + x^4 + x^2 + 1, x^7 + x^4 + x^2 + x, x^7 + x^4 + x^2 + x + 1, x^7 + x^4 + x^3, x^7 + x^4 + x^3 + 1, x^7 + x^4 + x^3 + x, x^7 + x^4 + x^3 + x + 1, x^7 + x^4 + x^3 + x^2, x^7 + x^4 + x^3 + x^2 + 1, x^7 + x^4 + x^3 + x^2 + x, x^7 + x^4 + x^3 + x^2 + x + 1, x^7 + x^5, x^7 + x^5 + 1, x^7 + x^5 + x, x^7 + x^5 + x + 1, x^7 + x^5 + x^2, x^7 + x^5 + x^2 + 1, x^7 + x^5 + x^2 + x, x^7 + x^5 + x^2 + x + 1, x^7 + x^5 + x^3, x^7 + x^5 + x^3 + 1, x^7 + x^5 + x^3 + x, x^7 + x^5 + x^3 + x + 1, x^7 + x^5 + x^3 + x^2, x^7 + x^5 + x^3 + x^2 + 1, x^7 + x^5 + x^3 + x^2 + x, x^7 + x^5 + x^3 + x^2 + x + 1, x^7 + x^5 + x^4, x^7 + x^5 + x^4 + 1, x^7 + x^5 + x^4 + x, x^7 + x^5 + x^4 + x + 1, x^7 + x^5 + x^4 + x^2, x^7 + x^5 + x^4 + x^2 + 1, x^7 + x^5 + x^4 + x^2 + x, x^7 + x^5 + x^4 + x^2 + x + 1, x^7 + x^5 + x^4 + x^3, x^7 + x^5 + x^4 + x^3 + 1, x^7 + x^5 + x^4 + x^3 + x, x^7 + x^5 + x^4 + x^3 + x + 1, x^7 + x^5 + x^4 + x^3 + x^2, x^7 + x^5 + x^4 + x^3 + x^2 + 1, x^7 + x^5 + x^4 + x^3 + x^2 + x, x^7 + x^5 + x^4 + x^3 + x^2 + x + 1, x^7 + x^6, x^7 + x^6 + 1, x^7 + x^6 + x, x^7 + x^6 + x + 1, x^7 + x^6 + x^2, x^7 + x^6 + x^2 + 1, x^7 + x^6 + x^2 + x, x^7 + x^6 + x^2 + x + 1, x^7 + x^6 + x^3, x^7 + x^6 + x^3 + 1, x^7 + x^6 + x^3 + x, x^7 + x^6 + x^3 + x + 1, x^7 + x^6 + x^3 + x^2, x^7 + x^6 + x^3 + x^2 + 1, x^7 + x^6 + x^3 + x^2 + x, x^7 + x^6 + x^3 + x^2 + x + 1, x^7 + x^6 + x^4, x^7 + x^6 + x^4 + 1, x^7 + x^6 + x^4 + x, x^7 + x^6 + x^4 + x + 1, x^7 + x^6 + x^4 + x^2, x^7 + x^6 + x^4 + x^2 + 1, x^7 + x^6 + x^4 + x^2 + x, x^7 + x^6 + x^4 + x^2 + x + 1, x^7 + x^6 + x^4 + x^3, x^7 + x^6 + x^4 + x^3 + 1, x^7 + x^6 + x^4 + x^3 + x, x^7 + x^6 + x^4 + x^3 + x + 1, x^7 + x^6 + x^4 + x^3 + x^2, x^7 + x^6 + x^4 + x^3 + x^2 + 1, x^7 + x^6 + x^4 + x^3 + x^2 + x, x^7 + x^6 + x^4 + x^3 + x^2 + x + 1, x^7 + x^6 + x^5, x^7 + x^6 + x^5 + 1, x^7 + x^6 + x^5 + x, x^7 + x^6 + x^5 + x + 1, x^7 + x^6 + x^5 + x^2, x^7 + x^6 + x^5 + x^2 + 1, x^7 + x^6 + x^5 + x^2 + x, x^7 + x^6 + x^5 + x^2 + x + 1, x^7 + x^6 + x^5 + x^3, x^7 + x^6 + x^5 + x^3 + 1, x^7 + x^6 + x^5 + x^3 + x, x^7 + x^6 + x^5 + x^3 + x + 1, x^7 + x^6 + x^5 + x^3 + x^2, x^7 + x^6 + x^5 + x^3 + x^2 + 1, x^7 + x^6 + x^5 + x^3 + x^2 + x, x^7 + x^6 + x^5 + x^3 + x^2 + x + 1, x^7 + x^6 + x^5 + x^4, x^7 + x^6 + x^5 + x^4 + 1, x^7 + x^6 + x^5 + x^4 + x, x^7 + x^6 + x^5 + x^4 + x + 1, x^7 + x^6 + x^5 + x^4 + x^2, x^7 + x^6 + x^5 + x^4 + x^2 + 1, x^7 + x^6 + x^5 + x^4 + x^2 + x, x^7 + x^6 + x^5 + x^4 + x^2 + x + 1, x^7 + x^6 + x^5 + x^4 + x^3, x^7 + x^6 + x^5 + x^4 + x^3 + 1, x^7 + x^6 + x^5 + x^4 + x^3 + x, x^7 + x^6 + x^5 + x^4 + x^3 + x + 1, x^7 + x^6 + x^5 + x^4 + x^3 + x^2, x^7 + x^6 + x^5 + x^4 + x^3 + x^2 + 1, x^7 + x^6 + x^5 + x^4 + x^3 + x^2 + x, x^7 + x^6 + x^5 + x^4 + x^3 + x^2 + x + 1]"}︡{"stdout":"\n"}︡{"done":true}
︠704cca1e-c2d3-4f9e-bc0d-17555b19c435s︠

︡a723eb96-eb72-4d50-88cd-5b51f9ffc436︡{"done":true}
︠c7b7a52e-ad3b-4baf-b937-2a005463c0fa︠

︡b32851ee-48b6-4c03-96d7-6eca3ab34518︡
︠3d2519e1-cf44-43ec-a44c-8c7c11229defs︠

︡b970d64a-e80e-4462-8f60-2785a10ce9a1︡{"done":true}
︠e01a7452-f453-42da-87e4-b1dcd3290d02︠














